package org.fuseri.moduleexercise.answer;

import jakarta.persistence.EntityNotFoundException;
import jakarta.validation.Valid;
import jakarta.validation.constraints.NotBlank;
import org.fuseri.model.dto.exercise.AnswerCreateDto;
import org.fuseri.model.dto.exercise.AnswerDto;
import org.fuseri.model.dto.exercise.AnswersCreateDto;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.server.ResponseStatusException;

import java.util.List;

/**
 * Represent a REST API controller for answers
 * Handle HTTP requests related to answers
 */
@RestController
@RequestMapping("/answers")
public class AnswerController {

    private final AnswerFacade facade;

    @Autowired
    public AnswerController(AnswerFacade facade) {
        this.facade = facade;
    }

    /**
     * Retrieve a list of AnswerDto objects which belong to question with questionId
     *
     * @param questionId the ID of the question for which to retrieve answers
     * @return a List of AnswerDto objects
     */
    @GetMapping("/{question-id}")
    public List<AnswerDto> findAllByQuestionId(@NotBlank @PathVariable("question-id") String questionId) {
        return facade.findAllByQuestionId(questionId);
    }

    /**
     * Create a new answer for the given question ID
     *
     * @param dto the AnswerCreateDto object containing information about the answer to create
     * @return an AnswerDto object representing the newly created answer
     * @throws ResponseStatusException if the question ID specified in the dto does not exist
     */
    @PostMapping
    public List<AnswerDto> createMultiple(@Valid @RequestBody AnswersCreateDto dto) {
        try {
            return facade.createMultiple(dto);
        } catch (EntityNotFoundException e) {
            throw new ResponseStatusException(HttpStatus.NOT_FOUND, e.getMessage());
        }
    }

    /**
     * Update an answer
     *
     * @param id  of answer to update
     * @param dto dto with updated answer information
     * @throws ResponseStatusException if the question id specified in the AnswerCreateDto dto does not exist
     */
    @PutMapping("/{id}")
    public AnswerDto update(@NotBlank @PathVariable String id, @Valid @RequestBody AnswerCreateDto dto) {
        try {
            return facade.update(id, dto);
        } catch (EntityNotFoundException e) {
            throw new ResponseStatusException(HttpStatus.NOT_FOUND,
                    e.getMessage());
        }
    }

    /**
     * Delete answer with the given id
     *
     * @param id of answer to delete
     * @throws ResponseStatusException if answer with specified id does not exist
     */
    @DeleteMapping("/{id}")
    public void delete(@NotBlank @PathVariable String id) {
        try {
            facade.delete(id);
        } catch (EntityNotFoundException e) {
            throw new ResponseStatusException(HttpStatus.NOT_FOUND, e.getMessage());
        }
    }
}
