package org.fuseri.moduleexercise.question;

import org.fuseri.model.dto.common.Result;
import org.fuseri.model.dto.exercise.QuestionCreateDto;
import org.fuseri.model.dto.exercise.QuestionDto;
import org.fuseri.model.dto.exercise.QuestionUpdateDto;
import org.fuseri.moduleexercise.answer.Answer;
import org.fuseri.moduleexercise.answer.AnswerMapper;
import org.fuseri.moduleexercise.answer.AnswerService;
import org.fuseri.moduleexercise.exercise.Exercise;
import org.fuseri.moduleexercise.exercise.ExerciseService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.stereotype.Service;

import java.util.HashSet;
import java.util.List;

/**
 * Represent facade for managing questions
 * Provide simplified interface for manipulating with questions
 */
@Service
public class QuestionFacade {
    private final QuestionService questionService;
    private final ExerciseService exerciseService;
    private final AnswerService answerService;
    private final QuestionMapper questionMapper;
    private final AnswerMapper answerMapper;

    @Autowired
    public QuestionFacade(
            QuestionService questionService, ExerciseService exerciseService,
            AnswerService answerService, QuestionMapper questionMapper,
            AnswerMapper answerMapper) {
        this.questionService = questionService;
        this.exerciseService = exerciseService;
        this.answerService = answerService;
        this.questionMapper = questionMapper;
        this.answerMapper = answerMapper;
    }

    /**
     * Find a question by ID.
     *
     * @param id the ID of the question to find
     * @return a QuestionDto object representing the found question
     */
    public QuestionDto find(String id) {
        var a = questionService.find(id);
        return questionMapper.toDto(a);
    }

    /**
     * Find questions by exercise ID and return them in a paginated format
     *
     * @param exerciseId the ID of the exercise to find questions for
     * @param page       the page number of the questions to retrieve
     * @return a Result object containing a list of QuestionDto objects and pagination information
     */
    public Result<QuestionDto> findByExerciseId(String exerciseId, int page) {
        Page<Question> questions = questionService.findByExerciseId(exerciseId, page);
        return questionMapper.toResult(questions);
    }

    /**
     * Create a new question
     *
     * @param dto a QuestionCreateDto object representing the new question to add
     * @return a QuestionDto object representing the added question
     */
    public QuestionDto create(QuestionCreateDto dto) {
        Question question = questionMapper.fromCreateDto(dto);

        Exercise exercise;
        exercise = exerciseService.find(question.getExerciseId());

        exercise.getQuestions().add(question);
        question.setExerciseId(exercise.getId());

        var answerDtos = dto.getAnswers();
        var answers = new HashSet<Answer>();
        for (var answerDto : answerDtos) {
            Answer answer = answerMapper.fromCreateDto(answerDto);
            answer = answerService.create(answer);
            answers.add(answer);
        }

        question.setAnswers(answers);
        var createdQuestion = questionService.create(question);

        for (var answer : answers) {
            answer.setQuestionId(createdQuestion.getId());
        }

        return questionMapper.toDto(createdQuestion);
    }

    /**
     * Update question
     *
     * @param dto dto of updated question with correct id
     * @return dto of updated question
     */
    public QuestionDto update(String id, QuestionUpdateDto dto) {
        Question question = questionMapper.fromUpdateDto(dto);
        question.setId(id);
        List<Answer> questionAnswers = answerService.findAllByQuestionId(id);
        question.setAnswers(new HashSet<>(questionAnswers));
        Question updatedQuestion = questionService.update(question);
        return questionMapper.toDto(updatedQuestion);
    }

    /**
     * Delete question
     *
     * @param id of qustion to delete
     */
    public void delete(String id) {
        var question = questionService.find(id);
        for (var answer : question.getAnswers()) {
            answerService.delete(answer.getId());
        }
        questionService.delete(id);
    }
}
