package org.fuseri.modulelanguageschool.course;

import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import jakarta.validation.Valid;
import org.fuseri.model.dto.course.CourseCreateDto;
import org.fuseri.model.dto.course.CourseDto;
import org.fuseri.model.dto.course.LanguageTypeDto;
import org.fuseri.model.dto.course.ProficiencyLevelDto;
import org.fuseri.model.dto.user.UserDto;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Sort;
import org.springframework.web.bind.annotation.*;

import java.util.List;

/**
 * This class represents a RESTful controller for Courses resources.
 * It handles incoming HTTP requests related to courses, and delegates them to the appropriate service method.
 */
@Api(tags = "Course Controller")
@RestController
@RequestMapping("/courses")
public class CourseController {

    private final CourseFacade courseFacade;

    @Autowired
    public CourseController(CourseFacade courseFacade) {
        this.courseFacade = courseFacade;
    }

    /**
     * Creates a new course.
     *
     * @param dto the CourseCreateDto containing the course data
     * @return the newly created CourseDto
     */
    @ApiOperation(value = "Create a new course")
    @PostMapping
    public CourseDto create(@Valid @RequestBody CourseCreateDto dto) {
        return courseFacade.create(dto);
    }

    /**
     * Retrieves a course by ID.
     *
     * @param id the ID of the course to retrieve
     * @return the CourseDto for the specified ID
     */
    @ApiOperation(value = "Retrieve a course by ID")
    @GetMapping("/find/{id}")
    public CourseDto find(@PathVariable Long id) {
        return courseFacade.findById(id);
    }

    /**
     * Retrieves a paginated list of courses
     *
     * @param page the page number to retrieve
     * @return the Result containing the requested page of CourseDtos
     */
    @ApiOperation(value = "Retrieve a paginated list of courses")
    @GetMapping("/findAll")
    public Page<CourseDto> findAll(@RequestParam int page) {
        return courseFacade.findAll(PageRequest.of(page, 10, Sort.by(Sort.Direction.ASC, "name")));
    }

    /**
     * Retrieves a paginated list of courses of a given language
     *
     * @param lang the language to find courses of
     * @return the Result containing the requested page of CourseDtos
     */
    @ApiOperation(value = "Retrieve a paginated list of courses of a given language")
    @GetMapping("/findAllByLang")
    public List<CourseDto> findAll(@RequestParam LanguageTypeDto lang) {
        return courseFacade.findAll(lang);
    }

    /**
     * Retrieves a paginated list of courses of a given language and proficiency
     *
     * @param lang the language to find courses of
     * @param prof the proficiency of the language
     * @return the Result containing the requested page of CourseDtos
     */
    @ApiOperation(value = "Retrieve a paginated list of courses of a given language and proficiency")
    @GetMapping("/findAllByLangProf")
    public List<CourseDto> findAll(@RequestParam LanguageTypeDto lang,
                                   @RequestParam ProficiencyLevelDto prof) {
        return courseFacade.findAll(lang, prof);
    }

    /**
     * Updates an existing course.
     *
     * @param id  the ID of the course to update
     * @param dto the CourseCreateDto containing the updated course data
     * @return the updated CourseDto
     */
    @ApiOperation(value = "Update an existing course")
    @PutMapping("/update/{id}")
    public CourseDto update(@PathVariable Long id, @Valid @RequestBody CourseCreateDto dto) {
        return courseFacade.update(id, dto);
    }

    /**
     * Deletes a course by ID.
     *
     * @param id the ID of the course to delete
     */
    @ApiOperation(value = "Delete a course by ID")
    @DeleteMapping("/delete/{id}")
    public void delete(@PathVariable Long id) {
        courseFacade.delete(id);
    }


    /**
     * Adds student to the existing course
     *
     * @param id      id of course to update
     * @param student UserDto for the student
     * @return the CourseDto representing the updated course
     */
    @ApiOperation(value = "Add student to the existing course")
    @PatchMapping("/enrol/{id}")
    public CourseDto enrol(@PathVariable Long id, @RequestParam Long studentId) {
        return courseFacade.enrol(id, studentId);
    }

    /**
     * Removes student from the existing course
     *
     * @param id      id of lecture to update
     * @param student UserDto for the student
     * @return the CourseDto representing the updated course
     */
    @ApiOperation(value = "Remove student from the existing course")
    @PatchMapping("/expel/{id}")
    public CourseDto expel(@PathVariable Long id, @RequestBody UserDto student) {
        return courseFacade.expel(id, student);
    }

}
